@@ -9,22 +9,22 @@ from __future__ import unicode_literals
 
 import collections
 
-from future.utils import bytes_to_native_str
 import caffe2
 from caffe2.python import core, workspace
 from caffe2.proto import caffe2_pb2
 import caffe2.python.utils
 import numpy as np
-from onnx import onnx_pb2, checker
-from onnx.onnx_pb2 import GraphProto, TensorProto, AttributeProto
+from onnx import checker, GraphProto, TensorProto, AttributeProto
 import onnx.numpy_helper
 import onnx.defs
-from onnx.backend.base import Backend, BackendRep, Device, DeviceType, namedtupledict
+from onnx.backend.base import Backend, Device, DeviceType, namedtupledict
 
 from onnx_caffe2.workspace import Workspace
 from onnx_caffe2.backend_rep import Caffe2Rep
 from onnx_caffe2.helper import dummy_name
 
+import warnings
+
 
 def get_device_option(device):
     m = {DeviceType.CPU: caffe2_pb2.CPU,
@@ -46,7 +46,8 @@ class OnnxAttributes(dict):
 
     def caffe2(self, kmap=lambda k: k):
         for k, v in self.items():
-            yield caffe2.python.utils.MakeArgument(kmap(k), v)
+            if kmap(k) != '':
+                yield caffe2.python.utils.MakeArgument(kmap(k), v)
 
 
 # TODO: Move this into ONNX main library
@@ -97,25 +98,44 @@ class OnnxNode(object):
 
 class Caffe2Backend(Backend):
 
+    # The greatest version of the ONNX operator set which we are aware of.
+    # Models whose version is larger than this will cause us to emit a warning
+    # that we are attempting to translate on a "best effort" basis.
+    #
+    # If you increase this, make SURE you cross-reference all BC-breaking
+    # changes from one version to the next, and any that you did not
+    # implement, mark as broken in _broken_operators
+    _known_opset_version = 2
+
+    # This dictionary will record operators which are KNOWN to be
+    # broken, so we give a good error message rather than do something
+    # bogus and then fail.
+    _broken_operators = {
+        # 'BrokenOp': version_it_was_broken_in
+    }
+
     # Operators that are different between Caffe2 and
     # ONNX but only in their name.
     # In most cases, this should be empty - as the effort of ONNX is
     # to unify the operator definitions.
     _renamed_operators = {
-        'Dot':                  'MatMul',
-        'Caffe2ConvTranspose':  'ConvTranspose',
-        'GlobalMaxPool':        'MaxPool',
-        'GlobalAveragePool':    'AveragePool',
-        'Pad':                  'PadImage',
-        'Neg':                  'Negative',
-        'BatchNormalization':   'SpatialBN',
+        'Caffe2ConvTranspose':   'ConvTranspose',
+        'GlobalMaxPool':         'MaxPool',
+        'GlobalAveragePool':     'AveragePool',
+        'Pad':                   'PadImage',
+        'Neg':                   'Negative',
+        'BatchNormalization':    'SpatialBN',
+        'InstanceNormalization': 'InstanceNorm',
+        'MatMul':                'BatchMatMul',
+        'Upsample':              'ResizeNearest',
+        'Equal':                 'EQ',
     }
 
     _global_renamed_attrs = {'kernel_shape': 'kernels'}
     _per_op_renamed_attrs = {
         'Squeeze':              {'axes': 'dims'},
         'Transpose':            {'perm': 'axes'},
-        'Pad':                  {'paddings': 'pads'},
+        'Upsample':             {'mode': ''},
     }
 
     # operators whose behavior is different beyond renaming
@@ -136,15 +156,11 @@ class Caffe2Backend(Backend):
         'Slice': '_create_slice',
     }
 
+    # NB: By default, you will use the LATEST definition of the operator,
+    # so this interface MAY make BC-breaking changes.  Specify an
+    # opset_version if you don't want this to version.
     @classmethod
-    def supports_device(cls, device):
-        device = Device(device)
-        if device.type == DeviceType.CUDA:
-            return workspace.has_gpu_support
-        return True
-
-    @classmethod
-    def run_node(cls, node, inputs, device='CPU'):
+    def run_node(cls, node, inputs, device='CPU', opset_version=_known_opset_version):
         super(Caffe2Backend, cls).run_node(node, inputs, device)
 
         device_option = get_device_option(Device(device))
@@ -153,14 +169,16 @@ class Caffe2Backend(Backend):
                 for key, value in inputs.items():
                     workspace.FeedBlob(key, value)
             else:
-                assert(len(node.input) == len(inputs))
+                assert len(node.input) == len(inputs), "{}: expected {} but got {}".format(
+                    node.op_type, len(node.input), len(inputs))
                 for key, value in zip(node.input, inputs):
                     workspace.FeedBlob(key, value)
 
             cls._inplace_rewrite([node])
-            ops = cls._onnx_node_to_caffe2_op(node)
+            ops = cls._onnx_node_to_caffe2_op(node, opset_version or cls._known_opset_version)
             for op in ops:
-                workspace.RunOperatorOnce(op)
+                op.device_option.CopyFrom(device_option)
+            workspace.RunOperatorsOnce(ops)
             output_values = [workspace.FetchBlob(name) for name in node.output]
             return namedtupledict('Outputs', node.output)(*output_values)
 
@@ -204,7 +222,7 @@ class Caffe2Backend(Backend):
             c2_values.ints.extend(tensor2list(onnx_tensor))
         elif onnx_tensor.data_type == TensorProto.STRING:
             c2_op.type = 'GivenTensorStringFill'
-            c2_values.strings.extend(tensor.string_data)
+            c2_values.strings.extend(onnx_tensor.string_data)
         else:
             raise RuntimeError(
                 "unrecognized tensor type {}".format(onnx_tensor.data_type))
@@ -218,12 +236,12 @@ class Caffe2Backend(Backend):
         return c2_op
 
     @classmethod
-    def _create_constant(cls, n):
+    def _create_constant(cls, n, opset_version):
         assert len(n.outputs) == 1
         return cls._create_tensor_filling_op(n.attrs["value"], n.outputs[0])
 
     @classmethod
-    def _create_gemm(cls, n):
+    def _create_gemm(cls, n, opset_version):
         (A, B, C) = n.inputs
         (Y,) = n.outputs
         alpha = n.attrs.get('alpha', 1.)
@@ -261,29 +279,30 @@ class Caffe2Backend(Backend):
         return ops
 
     @classmethod
-    def _create_pad(cls, n):
-        paddings = n.attrs['paddings']
-        if not (len(paddings) == 8 and
+    def _create_pad(cls, n, opset_version):
+        if opset_version < 2:
+            pads = n.attrs['paddings']
+        else:
+            pads = n.attrs['pads']
+        if not (len(pads) == 8 and
                 # first two dim is for batch and channel
-                set(paddings[:4]) == {0}):
-            raise ValueError('Caffe2 only supports padding 2D Tensor')
-        t, l, b, r = paddings[4:]
-        paddings[:] = [t, b, l, r]
-
-        return cls._common_onnx_node_to_caffe2_op(n)
+                set(pads[:2] + pads[4:6]) == {0}):
+            raise ValueError('Caffe2 only supports padding 2D Tensor, whereas padding is ' + str(pads))
+        pads[:] = pads[2:4] + pads[6:8]
+        return cls._common_onnx_node_to_caffe2_op(n, opset_version)
 
     @classmethod
-    def _create_concat(cls, n):
+    def _create_concat(cls, n, opset_version):
         # TODO: Caffe2 Concat has an extra output. It should be only
         # used when doing training, so we should change Caffe2 to allow
         # 1 output.
-        op = cls._common_onnx_node_to_caffe2_op(n)
+        op = cls._common_onnx_node_to_caffe2_op(n, opset_version)
         assert len(op.output) == 1
         op.output.append(dummy_name())
         return op
 
     @classmethod
-    def _create_optimized_rnn(cls, n):
+    def _create_optimized_rnn(cls, n, opset_version):
         # TODO: we cheat and rely on the fact that ONNX weight layout matches
         # CuDNN's. Properly we should extract the weight tensor and invoke
         # RecurrentParamSet exposed by C2
@@ -307,67 +326,109 @@ class Caffe2Backend(Backend):
         return op
 
     @classmethod
-    def _create_slice(cls, n):
-        op = cls._common_onnx_node_to_caffe2_op(n)
-        data, axes, orig_starts, orig_ends = op.input
+    def _create_slice(cls, n, opset_version):
+        op = cls._common_onnx_node_to_caffe2_op(n, opset_version)
+        args = {arg.name: arg for arg in op.arg}
+        starts_vals = np.array(
+            args.pop('starts').ints, dtype=np.int64).tolist()
+        ends_vals = np.array(
+            [i - 1 if i < 0 else i for i in args.pop('ends').ints],
+            dtype=np.int64).tolist()
+        if 'axes' in args:
+            axes_vals = np.array(
+                args.pop('axes').ints, dtype=np.int32).tolist()
+        else:
+            ndims = len(starts_vals)
+            axes_vals = np.array(range(ndims), dtype=np.int32).tolist()
+
+        data, = op.input
         ops = []
 
-        data_shape = dummy_name()
+        shape_tensor = dummy_name()
         ops.append(core.CreateOperator(
             'Shape',
             [data],
-            [data_shape]
+            [shape_tensor]
         ))
 
-        tmp_starts = dummy_name()
-        starts = dummy_name()
+        axes_tensor = dummy_name()
         ops.extend([
+            core.CreateOperator(
+                'GivenTensorIntFill',
+                [],
+                [axes_tensor],
+                shape=[len(axes_vals)],
+                values=axes_vals,
+            ),
+        ])
+
+        starts_vals_tensor = dummy_name()
+        starts_tensor = dummy_name()
+        casted_starts_tensor = dummy_name()
+        ops.extend([
+            core.CreateOperator(
+                'GivenTensorInt64Fill',
+                [],
+                [starts_vals_tensor],
+                shape=[len(starts_vals)],
+                values=starts_vals,
+            ),
             core.CreateOperator(
                 'ConstantFill',
-                [data_shape],
-                [tmp_starts],
+                [shape_tensor],
+                [starts_tensor],
                 dtype=caffe2_pb2.TensorProto.INT64,
                 value=0,
             ),
             core.CreateOperator(
                 'ScatterAssign',
-                [tmp_starts, axes, orig_starts],
-                [tmp_starts],
+                [starts_tensor, axes_tensor, starts_vals_tensor],
+                [starts_tensor],
             ),
             # Slice only accepts starts as int
             core.CreateOperator(
                 'Cast',
-                [tmp_starts],
-                [starts],
+                [starts_tensor],
+                [casted_starts_tensor],
                 to=caffe2_pb2.TensorProto.INT32,
             ),
         ])
 
-        tmp_ends = dummy_name()
-        ends = dummy_name()
+        ends_vals_tensor = dummy_name()
+        ends_tensor = dummy_name()
+        casted_ends_tensor = dummy_name()
         ops.extend([
+            core.CreateOperator(
+                'GivenTensorInt64Fill',
+                [],
+                [ends_vals_tensor],
+                shape=[len(ends_vals)],
+                values=ends_vals,
+            ),
             core.CreateOperator(
                 'ConstantFill',
-                [data_shape],
-                [tmp_ends],
+                [shape_tensor],
+                [ends_tensor],
                 dtype=caffe2_pb2.TensorProto.INT64,
                 value=-1,
             ),
             core.CreateOperator(
                 'ScatterAssign',
-                [tmp_ends, axes, orig_ends],
-                [tmp_ends],
+                [ends_tensor, axes_tensor, ends_vals_tensor],
+                [ends_tensor],
             ),
             # Slice only accepts ends as int
             core.CreateOperator(
                 'Cast',
-                [tmp_ends],
-                [ends],
+                [ends_tensor],
+                [casted_ends_tensor],
                 to=caffe2_pb2.TensorProto.INT32,
             ),
         ])
 
-        op.input[:] = [data, starts, ends]
+        op.input[:] = [data, casted_starts_tensor, casted_ends_tensor]
+        del op.arg[:]
+        op.arg.extend(args.values())
         ops.append(op)
 
         return ops
@@ -404,7 +465,7 @@ class Caffe2Backend(Backend):
     # differently.
 
     @classmethod
-    def _create_conv_pool_op_base(cls, n):
+    def _create_conv_pool_op_base(cls, n, opset_version):
         if n.op_type.startswith('Global'):
             n.attrs['global_pooling'] = 1
 
@@ -418,15 +479,28 @@ class Caffe2Backend(Backend):
                 # Caffe2 requires pads to be twice the size of kernels.
                 n.attrs['pads'] = pads * 2
 
-        return cls._common_onnx_node_to_caffe2_op(n)
+        return cls._common_onnx_node_to_caffe2_op(n, opset_version)
 
     @classmethod
-    def _create_reshape(cls, n):
-        c2_op = cls._common_onnx_node_to_caffe2_op(n)
+    def _create_reshape(cls, n, opset_version):
+        c2_op = cls._common_onnx_node_to_caffe2_op(n, opset_version)
         # Caffe2 has an extra output
         c2_op.output.append(dummy_name())
         return c2_op
 
+    @classmethod
+    def _direct_initialize_parameters(cls, initializer, ws, device_option):
+        for tp in initializer:
+            ws.FeedBlob(tp.name, onnx.numpy_helper.to_array(tp), device_option)
+
+    @classmethod
+    def _direct_initialize_inputs(cls, inputs, initialized, ws, device_option):
+        for value_info in inputs:
+            if value_info.name in initialized:
+                continue
+            shape = list(d.dim_value for d in value_info.type.tensor_type.shape.dim)
+            ws.FeedBlob(value_info.name, np.ones(shape), device_option)
+
     @classmethod
     def prepare(cls, model, device='CPU', **kwargs):
         '''
@@ -438,33 +512,74 @@ class Caffe2Backend(Backend):
         '''
         super(Caffe2Backend, cls).prepare(model, device, **kwargs)
 
-        init_net, predict_net = cls.onnx_graph_to_caffe2_net(model.graph)
-        predict_net.device_option.CopyFrom(get_device_option(Device(device)))
 
-        initialized = {init.name for init in model.graph.initializer}
-        uninitialized = [x for x in predict_net.external_input
-                         if not x in initialized]
+        opset_version = None
+        for imp in model.opset_import:
+            if not imp.HasField("domain") or imp.domain == "":
+                opset_version = imp.version
+                if imp.version > cls._known_opset_version:
+                    warnings.warn("This version of onnx-caffe2 targets ONNX operator set version {}, but the model we are trying to import uses version {}.  We will try to import it anyway, but if the model uses operators which had BC-breaking changes in the intervening versions, import will fail.".format(cls._known_opset_version, imp.version))
+            else:
+                warnings.warn("Unrecognized operator set {}".format(imp.domain))
+        if opset_version is None:
+            if model.ir_version >= 0x00000003:
+                raise RuntimeError("Model with IR version >= 3 did not specify ONNX operator set version (onnx-caffe2 requires it)")
+            else:
+                opset_version = 1
 
         ws = Workspace()
-        with ws, core.DeviceScope(predict_net.device_option):
-            workspace.RunNetOnce(init_net)
+        device_option = get_device_option(Device(device))
 
-        return Caffe2Rep(predict_net, ws, uninitialized)
+        # Directly load initializer data into blobs in workspace
+        cls._direct_initialize_parameters(
+            model.graph.initializer,
+            ws,
+            device_option,
+        )
+        # Need to pull this out before we delete model.graph.initializer
+        initialized = {init.name for init in model.graph.initializer}
+        initializer = model.graph.initializer[:]
+        # Delete the initializers so they aren't serialized
+        del model.graph.initializer[:]
+
+        cls._direct_initialize_inputs(
+            model.graph.input,
+            initialized,
+            ws,
+            device_option,
+        )
+        # Pull this out to manually add external inputs
+        external_inputs = model.graph.input[:]
+        del model.graph.input[:]
+
+        _, predict_net = cls.onnx_graph_to_caffe2_net(model.graph, device=device, opset_version=opset_version)
+        predict_net.external_input.extend(
+                value_info.name for value_info in external_inputs)
+
+        # Restore these so as not to mutate input
+        model.graph.initializer.extend(initializer)
+        model.graph.input.extend(external_inputs)
+
+        uninitialized = [x for x in predict_net.external_input
+                         if x not in initialized]
+
+        retval = Caffe2Rep(predict_net, ws, uninitialized)
+        return retval
 
     @classmethod
     # TODO: This method needs a refactor for clarity
-    def _onnx_node_to_caffe2_op(cls, node_def):
+    def _onnx_node_to_caffe2_op(cls, node_def, opset_version):
         if node_def.op_type in cls._special_operators:
             translator = getattr(cls, cls._special_operators[node_def.op_type])
         else:
             translator = cls._common_onnx_node_to_caffe2_op
-        ops = translator(OnnxNode(node_def))
+        ops = translator(OnnxNode(node_def), opset_version)
         if not isinstance(ops, collections.Iterable):
             ops = [ops]
         return ops
 
     @classmethod
-    def _common_onnx_node_to_caffe2_op(cls, onnx_node):
+    def _common_onnx_node_to_caffe2_op(cls, onnx_node, opset_version):
         """
         This translator performs the basic translation of ONNX nodes into
         Caffe2 operators.  Besides doing a straightforward marshalling from
@@ -484,6 +599,10 @@ class Caffe2Backend(Backend):
         c2_op.name = onnx_node.name
 
         onnx_op_type = onnx_node.op_type
+        broken_version = cls._broken_operators.get(onnx_op_type, float('Inf'))
+        if broken_version <= opset_version:
+            raise ValueError(
+                "Don't know how to translate op {} in ONNX operator set v{} (I only support prior to v{})".format(onnx_op_type, opset_version, broken_version))
         c2_op.type = cls._renamed_operators.get(onnx_op_type, onnx_op_type)
         if not core.IsOperator(c2_op.type):
             raise ValueError(
@@ -491,7 +610,7 @@ class Caffe2Backend(Backend):
 
         def kmap(k):
             if (onnx_op_type in cls._per_op_renamed_attrs and
-                k in cls._per_op_renamed_attrs[onnx_op_type]):
+                    k in cls._per_op_renamed_attrs[onnx_op_type]):
                 return cls._per_op_renamed_attrs[onnx_op_type][k]
             if k in cls._global_renamed_attrs:
                 return cls._global_renamed_attrs[k]
@@ -542,8 +661,22 @@ class Caffe2Backend(Backend):
             for output in graph_or_nodes.output:
                 output.name = renamed.get(output.name, output.name)
 
+    @staticmethod
+    def _all_names_in_graph(graph):
+        if graph is None:
+            return set()
+
+        names = set()
+        names.update(value_info.name for value_info in graph.input)
+        names.update(value_info.name for value_info in graph.output)
+        for node in graph.node:
+            names.update(node.input)
+            names.update(node.output)
+        return names
+
     @classmethod
-    def onnx_graph_to_caffe2_net(cls, graph_def):
+    def onnx_graph_to_caffe2_net(cls, graph_def, device="CPU", opset_version=_known_opset_version):
+        device_option = get_device_option(Device(device))
         cls._inplace_rewrite(graph_def)
         if graph_def.initializer:
             init_net = cls.onnx_initializer_to_caffe2_init_net(
@@ -553,10 +686,12 @@ class Caffe2Backend(Backend):
             init_net = caffe2_pb2.NetDef()
             initialized = set()
 
+        dummy_name(cls._all_names_in_graph(graph_def) | initialized)
+
         predict_net = caffe2_pb2.NetDef()
         predict_net.name = graph_def.name
         for node in graph_def.node:
-            predict_net.op.extend(cls._onnx_node_to_caffe2_op(node))
+            predict_net.op.extend(cls._onnx_node_to_caffe2_op(node, opset_version))
 
         predict_net.external_input.extend(
             value_info.name for value_info in graph_def.input)
@@ -573,6 +708,9 @@ class Caffe2Backend(Backend):
             op_def.type = 'GivenTensorFill'
 
             shape = list(d.dim_value for d in value_info.type.tensor_type.shape.dim)
+            # TODO: Putting this in the init net will make it run faster, but it
+            # causes some tests to fail...
+            # shape = (1,)
 
             shape_arg = op_def.arg.add()
             shape_arg.name = 'shape'
@@ -584,6 +722,10 @@ class Caffe2Backend(Backend):
 
             init_net.op.extend([op_def])
 
+        # Set the device option for the init_net and predict_net.
+        init_net.device_option.CopyFrom(device_option)
+        predict_net.device_option.CopyFrom(device_option)
+
         return init_net, predict_net
 
     @classmethod
@@ -610,4 +752,4 @@ run_node = Caffe2Backend.run_node
 
 run_model = Caffe2Backend.run_model
 
-supports_device = Caffe2Backend.supports_device
+supports_device = Caffe2Backend.supports_device  # noqa
